"""Module that contains the trajectory_component class."""
from pddl.pddl import Action
from typing import List

from sam_learner.sam_models.grounded_action import GroundedAction
from sam_learner.sam_models.parameter_binding import ParameterBinding
from sam_learner.sam_models.state import State


class TrajectoryComponent:
	"""Class describing a single step in a trajectory."""

	index: int
	previous_state: State
	grounded_action: GroundedAction
	next_state: State

	def __init__(self, index: int, previous_state: State, action: Action,
				 action_parameters_binding: List[ParameterBinding],
				 next_state: State):
		self.index = index
		self.previous_state = previous_state
		self.grounded_action = GroundedAction(action, action_parameters_binding)
		self.next_state = next_state

	def __str__(self):
		return f"Step number {self.index}: " \
			   f"previous state - {self.previous_state},\n" \
			   f"action applied - {self.grounded_action.activated_action_representation} " \
			   f"the next state - {self.next_state}."



"""Module to export plans using a remote solver server."""
import logging
import sys
from http import HTTPStatus
from pathlib import Path
from typing import NoReturn

import requests
from requests import Response

SOLVE_URL = 'http://solver.planning.domains/solve'
SOLVE_AND_VALIDATE_URL = 'http://solver.planning.domains/solve-and-validate'


class RemotePlansGenerator:
    """Class that uses an external service to generate plans for the learner algorithm.

    Attributes:
        logger: the logger of the class.
    """

    logger = logging.Logger

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def export_plan_from_response(self, problems_directory_path: str, problem_file_path: Path,
                                  response: Response) -> NoReturn:
        """Export the plan if exists into a solution file.

        :param problems_directory_path: the directory in which we export the output file to.
        :param problem_file_path: the path to the problem file (used to generate the solution
            file's name.
        :param response: the response that was returned from the solving server.
        """
        if response.status_code < HTTPStatus.BAD_REQUEST:
            response_data: dict = response.json()
            if "plan" not in response_data["result"]:
                return

            if response_data["result"]["val_status"] == "err":
                self.logger.debug(response_data["result"]["val_stdout"])
                return

            with open(Path(problems_directory_path, f"{problem_file_path.stem}_plan.solution"),
                      "wt") as plan_file:
                self.logger.debug("Solution Found!")
                plan_file.write(
                    '\n'.join([action["name"] for action in response_data["result"]["plan"]]))

    def generate_plans(self, domain_file_path: str, problems_directory_path: str, validate: bool = False) -> NoReturn:
        """Generates the plans using the solver that exists in the web.

        :param domain_file_path: the path to the domain file.
        :param problems_directory_path: the path to the directory containing the problems needed to solve.
        :param validate: whether or not to validate the input plans.
        """
        for file_path in Path(problems_directory_path).glob("*.pddl"):
            self.logger.info(f"Solving the problem {file_path.stem}")
            with open(domain_file_path, "rt") as domain_file, open(file_path, "rt") as problem_file:
                data = {"domain": domain_file.read(), "problem": problem_file.read()}
                url = SOLVE_AND_VALIDATE_URL if validate else SOLVE_URL
                response: Response = requests.post(url, verify=False, json=data)
                self.export_plan_from_response(problems_directory_path, file_path, response)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    RemotePlansGenerator().generate_plans(domain_file_path=sys.argv[1], problems_directory_path=sys.argv[2])



from .domain_export import DomainExporter
from .problem_export import ProblemExporter
from .effects_extractor import extract_effects
from .predicates_matcher import PredicatesMatcher
from .grounded_action_locator import ground_lifted_action, parse_plan_action_string, locate_lifted_action
from .random_walk_plan_generator import RandomWalkPlansGenerator
from .trajectory_generator import TrajectoryGenerator
from .remote_plan_generator import RemotePlansGenerator
from .trajectories_manager import TrajectorySerializationManager


"""Module that contains the SAM learner code for single agent problems."""
import logging
from pathlib import Path
from typing import List, Optional, NoReturn, Dict

from pddl.pddl import Domain, Action, Effect
from pyperplan import Parser

from sam_learner.core import PredicatesMatcher, extract_effects
from sam_learner.sam_models import GroundedAction, State, TrajectoryComponent, Trajectory, Mode


class SAMLearner:
	"""Class that represents the safe action model learner algorithm."""

	logger: logging.Logger
	working_directory_path: Path
	trajectories: List[Trajectory]
	learned_domain: Domain
	matcher: PredicatesMatcher
	known_actions: Dict[str, Action]

	def __init__(
			self, working_directory_path: Optional[str] = None, domain_file_name: str = "domain.pddl", mode: Mode = "production",
			domain: Optional[Domain] = None, known_actions: Dict[str, Action] = {}):
		self.logger = logging.getLogger(__name__)
		self.known_actions = known_actions
		if mode == "development":
			self.matcher = PredicatesMatcher(domain=domain)
			self.learned_domain = domain
			return

		self.working_directory_path = Path(working_directory_path)
		domain_path = self.working_directory_path / domain_file_name
		self.learned_domain = Parser(domain_path).parse_domain(read_from_file=True)
		self.learned_domain.actions = {}
		self.matcher = PredicatesMatcher(domain_path=str(domain_path))
		if known_actions is not None:
			self.learned_domain.actions = {
				name: action for name, action in known_actions.items()
			}

	def handle_action_effects(
			self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> Effect:
		"""Finds the effects generated from the previous and the next state on this current step.

		:param grounded_action: the grounded action that was executed according to the trajectory.
		:param previous_state: the state that the action was executed on.
		:param next_state: the state that was created after executing the action on the previous
			state.
		:return: the effect containing the add and del list of predicates.
		"""
		grounded_add_effects, grounded_del_effects = extract_effects(previous_state, next_state)
		action_effect = Effect()
		action_effect.addlist = action_effect.addlist.union(self.matcher.get_possible_literal_matches(
			grounded_action, grounded_add_effects))
		action_effect.dellist = action_effect.dellist.union(self.matcher.get_possible_literal_matches(
			grounded_action, grounded_del_effects))
		return action_effect

	def add_new_action(
			self, grounded_action: GroundedAction, previous_state: State, next_state: State) -> NoReturn:
		"""Create a new action in the domain.

		:param grounded_action: the grounded action that was executed according to the trajectory.
		:param previous_state: the state that the action was executed on.
		:param next_state: the state that was created after executing the action on the previous
			state.
		"""
		self.logger.info(f"Adding the action {grounded_action.activated_action_representation} "
						 f"to the domain.")
		new_action = Action(name=grounded_action.lifted_action_name,
							signature=grounded_action.lifted_signature,
							precondition=[],
							effect=None)

		# adding the preconditions each predicate is grounded in this stage.
		possible_preconditions = self.matcher.get_possible_literal_matches(grounded_action,
																		   previous_state.facts)
		new_action.precondition = possible_preconditions

		action_effect = self.handle_action_effects(grounded_action, previous_state, next_state)
		new_action.effect = action_effect
		self.learned_domain.actions[new_action.name] = new_action
		self.logger.debug(
			f"Finished adding the action {grounded_action.activated_action_representation}.")

	def update_action(self, grounded_action: GroundedAction, previous_state: State,
					  next_state: State) -> NoReturn:
		"""Create a new action in the domain.

		:param grounded_action: the grounded action that was executed according to the trajectory.
		:param previous_state: the state that the action was executed on.
		:param next_state: the state that was created after executing the action on the previous
			state.
		"""
		action_name = grounded_action.lifted_action_name
		self.logger.info(f"Updating the action - {action_name}")
		if action_name in self.known_actions:
			self.logger.debug(f"The action {action_name} is already known to the agent. Skipping!")
			return

		current_action: Action = self.learned_domain.actions[action_name]
		model_preconditions = current_action.precondition.copy()
		possible_preconditions = self.matcher.get_possible_literal_matches(
			grounded_action, previous_state.facts)

		for precondition in model_preconditions:
			if precondition not in possible_preconditions:
				current_action.precondition.remove(precondition)

		action_effect: Effect = self.handle_action_effects(
			grounded_action, previous_state, next_state)
		current_action.effect.addlist = current_action.effect.addlist.union(action_effect.addlist)
		current_action.effect.dellist = current_action.effect.dellist.union(action_effect.dellist)
		self.logger.debug(f"Done updating the action - {grounded_action.lifted_action_name}")

	def handle_single_trajectory_component(self, component: TrajectoryComponent) -> NoReturn:
		"""Handles a single trajectory component as a part of the learning process.

		:param component: the trajectory component that is being handled at the moment.
		"""
		previous_state = component.previous_state
		grounded_action = component.grounded_action
		next_state = component.next_state
		if grounded_action.lifted_action_name not in self.learned_domain.actions:
			self.add_new_action(grounded_action, previous_state, next_state)

		else:
			self.update_action(grounded_action, previous_state, next_state)

	def learn_action_model(self, trajectories: List[Trajectory]) -> Domain:
		"""Learn the SAFE action model from the input trajectories.

		:param trajectories: the list of trajectories that are used to learn the safe action model.
		:return: a domain containing the actions that were learned.
		"""
		# First making sure that the domain doesn't have any actions.
		self.logger.info("Starting to learn the action model!")
		for trajectory in trajectories:
			for component in trajectory:
				self.handle_single_trajectory_component(component)

		return self.learned_domain



"""module that represents a grounded action and its data."""
from pddl.pddl import Action, Type
from typing import List, Tuple

from sam_learner.sam_models.parameter_binding import ParameterBinding


class GroundedAction:
	"""Grounded action with bindings that represent the actual objects the action is applied on."""

	action: Action
	parameter_bindings: List[ParameterBinding]

	def __init__(self, action: Action, bindings: List[ParameterBinding]):
		self.action = action
		self.parameter_bindings = bindings

	@property
	def activated_action_representation(self) -> str:
		"""The signature of the grounded operator in the form of - (action o1 o2 o3 ... on)"""
		grounded_parameters = " ".join(
			[binding.grounded_object for binding in self.parameter_bindings])
		return f"({self.action.name} {grounded_parameters})"

	@property
	def grounded_signature(self) -> List[Tuple[str, Tuple[Type]]]:
		return [(binding.grounded_object, binding.parameter_types) for binding in
				self.parameter_bindings]

	@property
	def lifted_action_name(self) -> str:
		return self.action.name

	@property
	def lifted_signature(self) -> List[Tuple[str, Tuple[Type]]]:
		return self.action.signature



"""Module that represents a state."""
from typing import Tuple, List

from pddl.pddl import Domain, Problem

from sam_learner.sam_models.comparable_predicate import ComparablePredicate
from sam_learner.sam_models.parameter_binding import ParameterBinding


class StateLiteral:
	"""Represent the connection between a lifted and a grounded predicate."""

	fact: str
	domain: Domain

	def __init__(self, fact: str, domain: Domain):
		self.fact = fact
		self.domain = domain

	@classmethod
	def generate_problem_literal(cls, fact: str, domain: Domain, problem: Problem) -> Tuple[
		ComparablePredicate, List[ParameterBinding]]:
		"""Lift the grounded fact.

		:param fact: the fact to lift.
		:param domain: the domain containing all the data about the world.
		:param problem: the problem object containing the information about the concrete objects.
		:return: the mapping between the grounded fact and the lifted predicate.
		"""
		fact_data = fact.strip("()").split(" ")
		predicate_name = fact_data[0]
		predicate_objects = fact_data[1:]
		domain_predicate: ComparablePredicate = domain.predicates[predicate_name]
		bindings = []
		for (parameter_name, _), object_name in zip(domain_predicate.signature, predicate_objects):
			bindings.append(ParameterBinding(parameter_name,
											 (problem.objects[object_name],),
											 object_name))

		return domain_predicate, bindings

	@staticmethod
	def ground(grounding_data: Tuple[ComparablePredicate, List[ParameterBinding]]) -> str:
		"""ground lifted predicate data back to its fact form.

		:param grounding_data: the data of the object to ground.
		:return: the fact as a string form.
		"""
		predicate, bindings = grounding_data
		return f"({predicate.name} [{str(binding for binding in bindings)}])"


class State:
	"""Represent a state containing lifted predicates and their binding."""

	facts: List[ComparablePredicate]
	domain: Domain

	def __init__(self, facts: List[ComparablePredicate], domain: Domain):
		self.domain = domain
		self.facts = facts

	@classmethod
	def generate_problem_state(cls, facts: frozenset, domain: Domain, problem: Problem) -> List[
		Tuple[ComparablePredicate, List[ParameterBinding]]]:
		"""Lift the grounded facts to and returns the lifted facts and their object binding."""
		return [StateLiteral.generate_problem_literal(fact, domain, problem) for fact in facts]

	def __str__(self):
		return str(self.facts)

	def __eq__(self, other):
		return self.facts == other.facts

	def ground_facts(self) -> frozenset:
		"""Ground the facts to a form that the planner can use.

		:returns the grounded facts as a frozen set.
		"""
		grounded_facts = set()
		for fact in self.facts:
			if len(fact.signature) == 0:
				grounded_facts.add(f"({fact.name})")

			else:
				grounded_facts.add(f"({fact.name} {' '.join(item[0] for item in fact.signature)})")

		return grounded_facts



"""Matches predicates to their corresponding actions based on the common types."""
import copy
import logging
from typing import NoReturn, List, Tuple, Optional

from pddl.parser import Parser
from pddl.pddl import Type, Domain

from sam_learner.sam_models.comparable_predicate import ComparablePredicate
from sam_learner.sam_models.grounded_action import GroundedAction


def validate_no_duplicates(tested_list: List[str]) -> NoReturn:
	"""Validate that the predicate has only one possible match in the literal.

	:param tested_list: the list to test for duplicates.
	"""
	contains_duplicates = len(set(tested_list)) != len(tested_list)
	if contains_duplicates:
		raise ValueError(f"No duplications allowed! The predicates received - {tested_list}")


class PredicatesMatcher:
	"""Class that matches predicates according to the needed properties in the learning process."""

	matcher_domain: Domain
	logger: logging.Logger

	def __init__(self, domain_path: Optional[str] = None, domain: Optional[Domain] = None):
		self.logger = logging.getLogger(__name__)
		assert not (domain_path and domain)
		if domain_path is not None:
			self.matcher_domain = Parser(domain_path).parse_domain(read_from_file=True)

		if domain is not None:
			self.matcher_domain = Domain(
				name=domain.name,
				types=domain.types,
				predicates={name: ComparablePredicate(p.name, p.signature) for name, p in domain.predicates.items()},
				actions={},
				constants={name: constant for name, constant in domain.constants.items()}
			)

	def search_for_parameters_in_constants(
			self, possible_precondition: ComparablePredicate, grounded_predicate: ComparablePredicate) -> NoReturn:
		"""

		:param possible_precondition:
		:param grounded_predicate:
		:return:
		"""
		predicate_objects = [signature_item[0] for signature_item in grounded_predicate.signature]
		domain_predicate = self.matcher_domain.predicates[grounded_predicate.name]
		self.logger.debug("Searching for the predicate's parameters in the constants.")
		for index, constant_name in enumerate(self.matcher_domain.constants):
			if constant_name not in predicate_objects:
				continue

			literal_object_index = predicate_objects.index(constant_name)
			if domain_predicate.signature[literal_object_index] != \
					possible_precondition.signature[literal_object_index]:
				self.logger.debug("The parameters was already found in the action's signature, Skipping.")
				continue

			possible_precondition.signature[literal_object_index] = (
				constant_name, self.matcher_domain.constants[constant_name])

	def match_predicate_to_action_literals(
			self, grounded_predicate: ComparablePredicate,
			grounded_signature: List[Tuple[str, Tuple[Type]]],
			lifted_signature: List[Tuple[str, Tuple[Type]]]) -> Optional[ComparablePredicate]:
		"""Match a literal to a possible lifted precondition for the input action.

		:param grounded_predicate: the grounded predicate that represent part of the previous state.
		:param grounded_signature: the signature of the action that contains the actual objects
			that the action was executed on.
		:param lifted_signature: the lifted signature of the action, is accessible from the
			trajectory.
		:return: a precondition, in case the action and the predicate contain matching objects,
			None otherwise.
		"""
		predicate_objects = [signature_item[0] for signature_item in grounded_predicate.signature]
		grounded_action_objects = [signature_item[0] for signature_item in grounded_signature]
		constants_names = []
		if len(self.matcher_domain.constants) > 0:
			constants_names = [name for name in self.matcher_domain.constants]

		validate_no_duplicates(predicate_objects)
		validate_no_duplicates(grounded_action_objects)

		if not set(predicate_objects).issubset(set(grounded_action_objects).union(constants_names)):
			self.logger.debug("The predicate does not match the action with the constants")
			return None

		domain_predicate = self.matcher_domain.predicates[grounded_predicate.name]
		possible_signature = [(item[0], item[1]) for item in domain_predicate.signature]
		possible_precondition: ComparablePredicate = ComparablePredicate(
			domain_predicate.name, possible_signature)
		for index, (action_object_name, object_types) in enumerate(grounded_signature):
			if action_object_name not in predicate_objects:
				continue

			literal_object_index = predicate_objects.index(action_object_name)
			parameter_name, parameter_types = lifted_signature[index]
			possible_precondition.signature[literal_object_index] = (
				parameter_name, parameter_types)

		self.search_for_parameters_in_constants(possible_precondition, grounded_predicate)
		return possible_precondition

	def get_possible_literal_matches(self, grounded_action: GroundedAction,
									 literals: List[ComparablePredicate]) -> List[ComparablePredicate]:
		"""Get a list of possible preconditions for the action according to the previous state.

		:param grounded_action: the grounded action that was executed according to the trajectory.
		:param literals: the list of literals that we try to match according to the action.
		:return: a list of possible preconditions for the action that is being executed.
		"""
		possible_matches = []
		lifted_signature = grounded_action.lifted_signature
		grounded_signature = grounded_action.grounded_signature
		for predicate in literals:
			try:
				matches = self.match_predicate_to_action_literals(
					predicate, grounded_signature, lifted_signature)

			except ValueError as error:
				self.logger.debug(f"When parsing {grounded_action.activated_action_representation}, "
								  f"with the predicate {str(predicate)} "
								  f"got the error {error}! proceeding!")
				matches = None

			if matches is None:
				continue

			possible_matches.append(copy.deepcopy(matches))

		return possible_matches



"""This module extends the predicate class to enable the comparison between two predicates."""
from typing import Tuple, List, Optional

from pddl.pddl import Predicate, Type


class ComparablePredicate(Predicate):
    """Class that extends the basic predicate to enable comparison."""

    name: str
    signature: List[Tuple[str, Tuple[Type]]]

    def __init__(self, name: Optional[str] = None,
                 signature: Optional[List[Tuple[str, Tuple[Type]]]] = None,
                 predicate: Optional[Predicate] = None):
        if predicate:
            super(ComparablePredicate, self).__init__(predicate.name, predicate.signature)

        else:
            super(ComparablePredicate, self).__init__(name, signature)

    @staticmethod
    def is_sub_type(this_type: Type, other_type: Type) -> bool:
        """Checks if a type a subtype of the other.

        :param this_type: the checked type.
        :param other_type: the type that is checked to se if the first is subtype of.
        :return: whether or not the first is a subtype of the other.
        """
        ancestors_type_names = [this_type.name]
        compared_type = this_type
        while compared_type.parent is not None:
            ancestors_type_names.append(compared_type.parent.name)
            compared_type = compared_type.parent

        return other_type.name in ancestors_type_names

    @staticmethod
    def extract_types(signature: List[Tuple[str, Tuple[Type]]]) -> List[Type]:
        """Extract the type of the object from the signature format.

        :param signature: the signature of the predicate.
        :return: the types that were extracted from the signature.
        """
        types = []
        for _, param_type in signature:
            if type(param_type) is tuple or type(param_type) is list:
                types.append(param_type[0])
            else:
                types.append(param_type)
        return types

    def __eq__(self, other: Predicate):
        self_signature_params_name = [name for name, _ in self.signature]
        other_signature_params_name = [name for name, _ in other.signature]
        return \
            (self.name == other.name and self_signature_params_name ==
             other_signature_params_name and
             all([self.is_sub_type(this_type, other_type) for this_type, other_type in
                  zip(self.extract_types(self.signature), self.extract_types(other.signature))]))

    def __str__(self):
        return f"{self.name}{str(self.signature)}".strip("\n")

    def __hash__(self):
        return hash(str(self))

    def __copy__(self):
        return ComparablePredicate(self.name, self.signature)



"""Extract the effects from the state based on the exclusion and inclusion rules."""
from typing import Tuple, List

from pddl.pddl import Predicate

from sam_learner.sam_models.comparable_predicate import ComparablePredicate
from sam_learner.sam_models.state import State


def extract_effects(
		previous_state: State, next_state: State) -> Tuple[List[Predicate], List[Predicate]]:
	"""Extract the effects of the action according to the two lemmas that we know.

	:param previous_state: the state that had been before the action was executed.
	:param next_state: the state after the action was executed.
	:return: the add effects and the del effects.
	"""
	prev_state_predicates = set([ComparablePredicate(predicate=predicate) for predicate in
								 previous_state.facts])
	next_state_predicates = set([ComparablePredicate(predicate=predicate) for predicate in
								 next_state.facts])
	add_effects = next_state_predicates.difference(prev_state_predicates)
	del_effects = prev_state_predicates.difference(next_state_predicates)
	return list(add_effects), list(del_effects)



"""Module to extract statistics for the action model learner."""
import csv
import logging
import sys
from pathlib import Path
from typing import NoReturn, List, Dict, Any

from pddl.parser import Parser
from pddl.pddl import Domain

from sam_learner import SAMLearner
from sam_learner.core import DomainExporter
from sam_learner.core.trajectories_manager import TrajectorySerializationManager
from sam_learner.model_validation.domain_comparison import validate_action_model
from sam_learner.sam_models.types import Trajectory

COLUMN_NAMES = [
	"domain_name",
	"num_trajectories",
	"num_domain_actions",
	"num_known_actions",
	"learned_actions_length",
	"domains_equal",
	"avg_trajectories_length",
	"learned_action_names",
	"error"]

LEARNED_ACTIONS_STATS_COLUMNS = [
	"domain_name",
	"num_trajectories",
	"learned_action_name",
	"learned_preconditions",
	"learned_add_effects",
	"learned_delete_effects"
]


def validate_no_redundant_actions(known_actions: List[str], learner: SAMLearner) -> NoReturn:
	"""Validate that the initial state of the domain does not contain any redundant actions.

	:param known_actions: the names of the actions that the agent already knows.
	:param learner: the learning algorithm.
	"""
	if known_actions is None:
		assert learner.learned_domain.actions == {}
	else:
		assert len(learner.learned_domain.actions) == len(known_actions)


class AMStatisticsExtractor:
	"""Class that is used to extract statistics about the action model learner.

	Attributes:
		logger: the logger for the class.
		working_dir_path: the path to the directory containing the domain and the relevant files.
		expected_domain: the domain that contains all of the actions and the preconditions and
			effects.
		domain_exporter: the class that is able to export the domain to a PDDL file.
	"""

	logger: logging.Logger
	workdir_path: str
	expected_domain: Domain
	domain_exporter: DomainExporter
	expected_domain_file_name: str
	trajectories_manager: TrajectorySerializationManager

	def __init__(self, working_dir_path: str, expected_domain_path: str):
		self.logger = logging.getLogger(__name__)
		self.workdir_path = working_dir_path
		self.expected_domain = Parser(expected_domain_path).parse_domain(read_from_file=True)
		self.expected_domain_file_name = f"{Path(expected_domain_path).stem}.pddl"
		self.domain_exporter = DomainExporter()
		self.trajectories_manager = TrajectorySerializationManager(
			workdir_path=Path(working_dir_path), domain_path=Path(expected_domain_path))

	def create_trajectories(self) -> List[Trajectory]:
		"""Create the trajectories for the learner.

		:return: the trajectories needed to learn the action model.
		"""
		self.logger.info("Creating the trajectories for the learner to use.")
		return self.trajectories_manager.create_trajectories()

	def run_learner(
			self, complete_statistics: List[Dict[str, any]], learned_actions_statistics: List[Dict[str, Any]],
			known_actions: List[str]) -> NoReturn:
		"""Runs the learner and accumulates the statistics from the run.

		:param complete_statistics: the list of statistics that is to be accumulated.
		:param learned_actions_statistics: the statistics objects containing information about the learned actions.
		:param known_actions: the actions that the agent already know and thus there is no need to learn them.
		"""
		available_trajectories = []
		domain_trajectories = self.create_trajectories()
		(Path(self.workdir_path) / "generated_domains").mkdir(exist_ok=True)
		for trajectory in domain_trajectories:
			statistics = {}
			available_trajectories.append(trajectory)
			known_actions_map = {name: self.expected_domain.actions[name] for name in known_actions}
			learner = SAMLearner(
				working_directory_path=self.workdir_path, known_actions=known_actions_map,
				domain_file_name=self.expected_domain_file_name)
			validate_no_redundant_actions(known_actions, learner)

			self.logger.debug(
				f"Trying to learn the action model with {len(available_trajectories)} trajectories.")
			try:
				learned_model = learner.learn_action_model(available_trajectories)
				print("Finished learning the action model from the trajectories!")
				self.write_statistics_data(available_trajectories, learned_model, statistics, known_actions)
				self.write_learned_actions_statistics(
					available_trajectories, learned_model, learned_actions_statistics, known_actions)
				self.domain_exporter.export_domain(
					learned_model,
					Path(f"{self.workdir_path}/generated_domains/learned_domain-"
						 f"{len(available_trajectories)}-trajectories.pddl"))
			except ValueError as error:
				self.write_learning_error(error, statistics)
				available_trajectories.pop()

			complete_statistics.append(statistics)

	def write_learning_error(self, error: Exception, statistics: dict) -> NoReturn:
		"""Write the appropriate statistics data in case that there is an error in the learning process.

		:param error: the error that was received while trying to learn the action model.
		:param statistics: the dictionary containing the statistics about the current run.
		"""
		self.logger.warning(error)
		statistics["domain_name"] = self.expected_domain.name
		statistics["error"] = error

	def write_learned_actions_statistics(
			self, available_trajectories: List[Trajectory], learned_model: Domain,
			learned_actions_statistics: List[Dict[str, Any]], known_actions: List[str]) -> NoReturn:
		"""Writes the statistics about the learned actions. This adds additional information about the learned actions.

		:param available_trajectories: the trajectories used in the learning process.
		:param learned_model: the action model that was learned.
		:param learned_actions_statistics: the dictionary containing the data about the learned actions.
		:param known_actions: the actions that the agent already knows.
		"""
		num_trajectories = len(available_trajectories)
		learned_actions = [
			action for action in learned_model.actions if action not in known_actions]
		for action_name in learned_actions:
			action_statistics = {
				"domain_name": self.expected_domain.name, "num_trajectories": num_trajectories,
				"learned_action_name": action_name
			}
			action_preconditions = learned_model.actions[action_name].precondition
			action_add_effects = list(learned_model.actions[action_name].effect.addlist)
			action_delete_effects = list(learned_model.actions[action_name].effect.dellist)
			action_statistics["learned_preconditions"] = ",".join([str(precond) for precond in action_preconditions])
			action_statistics["learned_add_effects"] = ",".join([str(add_effect) for add_effect in action_add_effects])
			action_statistics["learned_delete_effects"] = ",".join(
				[str(del_effect) for del_effect in action_delete_effects])
			learned_actions_statistics.append(action_statistics)

	def write_statistics_data(
			self, available_trajectories: List[Trajectory], learned_model: Domain, statistics: dict,
			known_actions: List[str]) -> NoReturn:
		"""Write the statistics of the current execution of the code.

		:param available_trajectories: the available trajectories for the current execution.
		:param learned_model: the domain that was learned with the action model.
		:param statistics: the dictionary that contains the statistics collected during the
			entire program's execution.
		:param known_actions: the actions that were known to the model before the learning process started.
		:return:
		"""
		num_trajectories = len(available_trajectories)
		statistics["domain_name"] = self.expected_domain.name
		statistics["num_trajectories"] = num_trajectories
		statistics["num_domain_actions"] = len(self.expected_domain.actions)
		statistics["num_known_actions"] = len(known_actions)
		statistics["learned_actions_length"] = len(learned_model.actions) - len(known_actions)
		statistics["domains_equal"] = validate_action_model(learned_model, self.expected_domain)
		statistics["avg_trajectories_length"] = sum([
			len(trajec) for trajec in available_trajectories]) / num_trajectories
		statistics["learned_action_names"] = "/".join([
			action for action in learned_model.actions if action not in known_actions])
		statistics["error"] = None

	def extract_action_model_statistics(self, stats_file_path: str, actions_stats_file_path: str,
										known_actions: List[str] = []) -> NoReturn:
		"""Extract the statistics and saves them into a CSV file for future usage.

		:param stats_file_path: the path to the file that will contain the statistics.
		:param actions_stats_file_path: the path to the file containing data about the learned actions.
		:param known_actions: the names of the actions that the agent already knows.
		"""
		domain_statistics = []
		learned_actions_statistics = []
		self.run_learner(domain_statistics, learned_actions_statistics, known_actions)
		with open(stats_file_path, 'wt', newline='') as domain_stats_csv_file, \
				open(actions_stats_file_path, 'wt', newline='') as actions_stats_csv_file:
			domain_data_writer = csv.DictWriter(domain_stats_csv_file, fieldnames=COLUMN_NAMES)
			actions_data_writer = csv.DictWriter(actions_stats_csv_file, fieldnames=LEARNED_ACTIONS_STATS_COLUMNS)
			domain_data_writer.writeheader()
			actions_data_writer.writeheader()
			for data in domain_statistics:
				domain_data_writer.writerow(data)

			for data in learned_actions_statistics:
				actions_data_writer.writerow(data)


if __name__ == '__main__':
	try:
		logging.basicConfig(level=logging.DEBUG)
		args = sys.argv
		workdir_path = args[1]
		complete_domain_path = args[2]
		statistics_output_path = args[3]
		actions_statistics_output_path = args[4]
		stats_extractor = AMStatisticsExtractor(workdir_path, complete_domain_path)
		stats_extractor.extract_action_model_statistics(statistics_output_path, actions_statistics_output_path)

	except Exception as e:
		print(e)



from .action_model_statistics_extractor import AMStatisticsExtractor
from .learned_domain_validator import DomainValidator, ValidationResult
from .domain_comparison import switch_comparisons_and_sort, validate_add_effects, validate_delete_effects, \
	validate_preconditions, validate_action_model



"""Exports a problem file to a PDDL file according to the input given."""
from pathlib import Path
from typing import NoReturn, List, Dict

from pddl.pddl import Predicate, Type, Problem


class ProblemExporter:
	"""Class that is able to export a domain to a correct PDDL file."""

	@staticmethod
	def write_objects(problem_objects: Dict[str, Type]) -> str:
		"""Writes the definitions of the types according to the PDDL file format.

		:param problem_objects: the objects that are available in the learned domain.
		:return: the formatted string representing the objects in the PDDL problem file.
		"""
		objects_str = "(:objects\n{objects_content}\n)\n"
		objects = []
		for object_name, pddl_type in problem_objects.items():
			objects.append(f"\t{object_name} - {pddl_type.name}")

		return objects_str.format(objects_content="\n".join(objects))

	@staticmethod
	def write_initial_state(initial_state: List[Predicate]) -> str:
		"""Writes the definitions of the types according to the PDDL file format.

		:param initial_state: the objects that are available in the learned domain.
		:return: the formatted string representing the state in the PDDL problem file.
		"""
		state_str = "(:init\n{state_content}\n)\n"
		predicates = ProblemExporter.extract_state_predicates(initial_state)
		return state_str.format(state_content="\n".join(predicates))

	@staticmethod
	def write_goal_state(goal_state: List[Predicate]) -> str:
		"""Writes the definitions of the types according to the PDDL file format.

		:param goal_state: the objects that are available in the learned domain.
		:return: the formatted string representing the state in the PDDL problem file.
		"""
		state_str = "(:goal\n\t(and\n{state_content}\t\n)\n)\n"
		predicates = ProblemExporter.extract_state_predicates(goal_state)
		return state_str.format(state_content="\n".join(predicates))

	@staticmethod
	def extract_state_predicates(state: List[Predicate]) -> List[str]:
		"""Extract the needed problem predicates for the PDDL file representation.

		:param state: the state to write in a PDDL format.
		:return: the strings of containing the state's data.
		"""
		predicates = []
		for predicate in state:
			predicate_objects = [obj[0] for obj in predicate.signature]
			objects_str = " ".join(predicate_objects)
			predicates.append(f"\t({predicate.name} {objects_str})")
		predicates = list(set(predicates))
		return predicates

	def export_problem(self, problem: Problem, export_path: Path) -> NoReturn:
		"""Export the domain object to a correct PDDL file.

		:param problem: the problem object to export to a PDDL file.
		:param export_path: the path to the file that the domain would be exported to.
		"""
		problem_objects = self.write_objects(problem.objects)
		initial_state = self.write_initial_state(problem.initial_state)
		goal_state = self.write_goal_state(problem.goal)
		problem_data = f"(define (problem {problem.name}) (:domain {problem.domain.name})\n" \
					   f"{problem_objects}\n" \
					   f"{initial_state}\n" \
					   f"{goal_state}\n)"

		with open(export_path, "wt") as export_problem_file:
			export_problem_file.write(problem_data)



"""This module is used to validate the correctness of the generated action model."""
from typing import Literal, Tuple, List

from pddl.pddl import Domain, Action

from sam_learner.sam_models.comparable_predicate import ComparablePredicate

ValidationType = Literal["precondition", "add-effect", "del-effect"]


def switch_comparisons_and_sort(expected_action: Action, generated_action: Action,
								validated_type: ValidationType) -> Tuple[List, List]:
	"""Creates the list of items to be compared and sorts them so that the comparison will be
	accurate.

	:param expected_action: the expected action that is in the complete domain.
	:param generated_action: the action that was learned using the learner algorithm.
	:param validated_type: the type of objects to compare.
	:return:
	"""
	if validated_type == "add-effect":
		generated_objects = generated_action.effect.addlist
		expected_objects = expected_action.effect.addlist
	elif validated_type == "precondition":
		generated_objects = generated_action.precondition
		expected_objects = expected_action.precondition
	else:
		generated_objects = generated_action.effect.dellist
		expected_objects = expected_action.effect.dellist

	generated_objects = [ComparablePredicate(predicate=pred) for pred in generated_objects]
	expected_objects = [ComparablePredicate(predicate=pred) for pred in expected_objects]
	generated_objects = sorted(generated_objects, key=lambda predicate: (predicate.name,
																		 str(predicate.signature)))
	expected_objects = sorted(expected_objects, key=lambda predicate: (predicate.name,
																	   str(predicate.signature)))
	return expected_objects, generated_objects


def validate_objects(generated_action: Action, expected_action: Action, validated_type: ValidationType) -> bool:
	"""Validate that the actions contain the same precondition / add effects / delete effects.

	:param generated_action: the action that was generated using the learning algorithm.
	:param expected_action: the actual action that is in the planning domain.
	:param validated_type: the type of object to test.
	:return: whether or not the items that were learned are the same.
	"""
	expected_objects, generated_objects = switch_comparisons_and_sort(
		expected_action, generated_action, validated_type)
	if len(generated_objects) == len(expected_objects) == 0:
		print(f"No {validated_type} in both the generated and the expected!")
		return True

	generated_textual = set([str(predicate) for predicate in generated_objects])
	expected_textual = set([str(predicate) for predicate in expected_objects])

	print(f"Generated {validated_type}: {generated_textual}")
	print(f"Expected {validated_type}: {expected_textual}")

	if len(generated_textual) != len(expected_textual):
		print(
			f"The length of the given objects does not match! Expected {len(expected_textual)} "
			f"but received {len(generated_textual)}")
		return False

	for generated_item, expected_item in zip(generated_objects, expected_objects):
		comparable_item = ComparablePredicate(predicate=generated_item)
		if comparable_item == expected_item:
			print(f"The {validated_type}s are either equal or the firsts if more precise.")
			return True

		return False


def validate_preconditions(generated_action: Action, expected_action: Action) -> bool:
	return validate_objects(generated_action, expected_action, "precondition")


def validate_add_effects(generated_action: Action, expected_action: Action) -> bool:
	return validate_objects(generated_action, expected_action, "add-effect")


def validate_delete_effects(generated_action: Action, expected_action: Action) -> bool:
	return validate_objects(generated_action, expected_action, "del-effect")


def validate_action_model(generated_domain: Domain, complete_domain: Domain) -> bool:
	"""This function validates whether or not the generated domain is the same as the complete one.

	:param generated_domain: the domain containing the learned action model.
	:param complete_domain: the actual domain that we compare to.
	:return: whether or not the action models are equal.
	"""
	if len(generated_domain.actions) != len(complete_domain.actions):
		print("Action models have different lengths!")
		print(f"Generated - {generated_domain.actions}")
		print(f"Expected - {complete_domain.actions}")
		return False

	for action_name, generated_action_data in generated_domain.actions.items():
		print(f"Checking the action - {action_name}")
		assert action_name in complete_domain.actions
		expected_action = complete_domain.actions[action_name]

		if (validate_preconditions(generated_action_data, expected_action) and
				validate_add_effects(generated_action_data, expected_action) and
				validate_delete_effects(generated_action_data, expected_action)):
			continue

		return False

	return True



"""represent a biding between a parameter and its respected object / other parameter name."""
from typing import Tuple

from pddl.pddl import Type


class ParameterBinding:
	"""Class representing the binding between a parameter and its respected objects."""

	parameter_name: str
	parameter_types: Tuple[Type]
	grounded_object: str

	def __init__(self, name: str, types: Tuple[Type], grounded_object: str):
		self.parameter_name = name
		self.parameter_types = types
		self.grounded_object = grounded_object

	def __str__(self):
		return f"('{self.grounded_object}', {self.parameter_types[0]})"

	def __repr__(self):
		return f"parameter - {self.parameter_name} with the types - {self.parameter_types} " \
			   f"contains the grounded object - {self.grounded_object}"

	def __eq__(self, other):
		return (self.parameter_name == other.parameter_name
				and str(self.parameter_types) == str(other.parameter_types)
				and self.grounded_object == other.grounded_object)

	def bind_parameter(self) -> Tuple[str, Type]:
		"""Bind the predicate's signature to the relevant object.

		:return: the grounded predicate signature.
		"""
		return self.grounded_object, self.parameter_types[0]



"""Module to handle the storing loading and updating of a trajectory."""
import logging
import os
import pickle
from pathlib import Path
from typing import Tuple, List, NoReturn, Optional

from pddl.pddl import Domain
from sam_learner.core import TrajectoryGenerator
from sam_learner.sam_models import Trajectory


class TrajectorySerializationManager:
	"""class that manages the serialization processes of the trajectories."""

	working_directory_path: Path
	logger: logging.Logger
	domain_path: Path

	def __init__(self, workdir_path: Path, domain_path: Path):
		self.working_directory_path = workdir_path
		self.domain_path = domain_path
		self.logger = logging.getLogger(__name__)

	def get_problem_and_solution_files(self) -> List[Tuple[Path, Path]]:
		"""Get the problem and the solution file paths from the working directory.

		:return: the paths to the problems and their respected plans.
		"""
		paths = []
		for solution_file_path in self.working_directory_path.glob("*.solution"):
			problem_file_name = solution_file_path.stem.split("_plan")[0]
			problem_path = self.working_directory_path / f"{problem_file_name}.pddl"
			paths.append((problem_path, solution_file_path))

		return paths

	def create_trajectories(self, serialization_file_name: Optional[str] = None) -> List[Trajectory]:
		"""Create the trajectories that will be used in the main algorithm.

		:return: the list of trajectories that will be used in the SAM algorithm.
		"""
		self.logger.info("Creating the trajectories for the algorithm.")
		trajectories = []
		if serialization_file_name is not None:
			stored_trajectories_path = self.working_directory_path / serialization_file_name

		else:
			stored_trajectories_path = self.working_directory_path / "saved_trajectories"

		if stored_trajectories_path.exists():
			return self.load_trajectories(stored_trajectories_path)

		for problem_path, plan_path in self.get_problem_and_solution_files():
			generator = TrajectoryGenerator(str(self.domain_path), str(problem_path))
			trajectories.append(generator.generate_trajectory(str(plan_path)))

		self.store_trajectories(stored_trajectories_path, trajectories)
		return trajectories

	def load_trajectories(self, stored_trajectories_path: Path) -> List[Trajectory]:
		"""Loads the trajectories from the trajectories file.

		:param stored_trajectories_path: the path to the files that stores the trajectories.
		:return: the loaded deserialized trajectories.
		"""
		self.logger.debug("Loading the trajectories from the file!")
		with open(stored_trajectories_path, "rb") as trajectories_file:
			return pickle.load(trajectories_file)

	def store_trajectories(self, stored_trajectories_path: Path, trajectories: List[Trajectory]) -> NoReturn:
		"""Store the trajectories in the trajectory file so that future runs of the algorithm would be faster.

		:param stored_trajectories_path: the path to the file that stores the trajectories.
		:param trajectories: the trajectories that are to be stored in the file.
		"""
		with open(stored_trajectories_path, "wb") as trajectories_file:
			self.logger.debug("Saving the created trajectories in a file for future usage.")
			pickle.dump(trajectories, trajectories_file)

	def update_stored_trajectories(self, trajectory: Trajectory, save_path: Optional[Path] = None) -> NoReturn:
		"""Serialize the new trajectory.

		:param trajectory: the trajectory to serialize.
		:param save_path: the path to the file that saves the trajectories.
		"""
		trajectories_path = self.working_directory_path / "saved_trajectories" if save_path is None else save_path
		with open(trajectories_path, "rb") as trajectories_file:
			trajectories = pickle.load(trajectories_file)

		trajectories.append(trajectory)
		self.store_trajectories(trajectories_path, trajectories)

	def delete_trajectories_file(self, trajectories_file_path: Optional[Path] = None) -> NoReturn:
		"""deletes the file that contains the saved trajectories.

		:param trajectories_file_path: the path to the trajectories file.
		"""
		trajectories_path = self.working_directory_path / "saved_trajectories" if \
			trajectories_file_path is None else trajectories_file_path
		os.remove(trajectories_path)



"""Module that converts a problem data and a plan to a trajectory."""
import logging
from typing import List, Tuple

import grounding
from pddl.pddl import Problem, Domain, Action, Predicate
from pyperplan import Parser
from task import Operator

from sam_learner.core.grounded_action_locator import parse_plan_action_string, locate_lifted_action, \
	ground_lifted_action
from sam_learner.sam_models.parameter_binding import ParameterBinding
from sam_learner.sam_models.state import State
from sam_learner.sam_models.trajectory_component import TrajectoryComponent
from sam_learner.sam_models.types import Trajectory


class TrajectoryGenerator:
	"""Class that generates trajectory out of a problem file and it's corresponding plan."""

	logger: logging.Logger
	domain: Domain
	problem: Problem

	def __init__(self, domain_path: str, problem_path: str):
		parser = Parser(domain_path, problem_path)
		self.logger = logging.getLogger(__name__)
		self.domain = parser.parse_domain(read_from_file=True)
		self.problem = parser.parse_problem(dom=self.domain, read_from_file=True)

	def _parse_plan_actions(self, plan_path: str) -> Tuple[List[str], List[Action]]:
		"""Parse the actions that exist in the plan file and extract the action objects with the
		object binding.

		:param plan_path: the path to the file containing the plan that solves the input problem.
		:return: both the grounded operator names and the actions to execute.
		"""
		self.logger.debug(f"Parsing the plan file in the path - {plan_path}")
		with open(plan_path, "rt") as plan_file:
			lines = plan_file.readlines()
			lines = [line.strip("\n") for line in lines]
			if "cost" in lines[-1]:
				lines.pop()  # remove the cost line from the plan.

		actions = []
		self.logger.debug("The plan contains the following grounded actions:")
		for index, line in enumerate(lines):
			self.logger.debug(line)
			action_name, action_params = parse_plan_action_string(line)
			if len(action_params) == 0:
				lines[index] = f"({action_name})"

			actions.append(locate_lifted_action(self.domain, action_name))

		return lines, actions

	def _create_state_parameter_matching(self, grounded_state: frozenset) -> State:
		"""Matches between the grounded and the lifted predicates.

		:param grounded_state: the grounded set of facts that represent the state.
		:return: a match between the grounded and the lifted predicates.
		"""
		lifted_state_data: List[Tuple[Predicate, List[ParameterBinding]]] = \
			State.generate_problem_state(grounded_state, self.domain, self.problem)
		grounded_predicates = []
		for predicate, bindings in lifted_state_data:
			signature = [binding.bind_parameter() for binding in bindings]
			grounded_predicates.append(Predicate(predicate.name, signature))

		return State(grounded_predicates, self.domain)

	def _create_single_trajectory_component(
			self, index: int, action: Action, op_name: str,
			operators: dict, previous_state: State) -> TrajectoryComponent:
		"""Create a single trajectory component by applying the action on the previous state.

		:param index: the index of the step that is being parsed currently.
		:param action: the lifted action that is to be executed on the state.
		:param op_name: the grounded operator's name.
		:param operators: the grounded operators that can apply an action on a state.
		:param previous_state: the previous state to be changed.
		:return: the trajectory component representing the current stage.
		"""
		grounded_action: Operator = operators[op_name]
		grounded_next_state_statements = grounded_action.apply(previous_state.ground_facts())
		next_state = self._create_state_parameter_matching(grounded_next_state_statements)
		_, action_objects = parse_plan_action_string(grounded_action.name)
		_, bindings = ground_lifted_action(action, action_objects)
		component = TrajectoryComponent(index=index,
										previous_state=previous_state,
										action=action,
										action_parameters_binding=bindings,
										next_state=next_state)
		self.logger.debug(f"Finished creating the component:\n{component}")
		return component

	def generate_trajectory(self, plan_path: str, should_return_partial_trajectory: bool = False) -> Trajectory:
		"""Generates a trajectory out of a problem file and a plan file.

		:param plan_path the path to the plan generated by a solver.
		:param should_return_partial_trajectory: whether or not to return a partial trajectory in case of failure.
		:return: the trajectory that represents the plan file.

		Note:
			the plan output should be in the form of: (load-truck obj23 tru2 pos2) ...
		"""
		self.logger.info(f"Generating a trajectory from the file: {self.problem}")
		trajectory = []
		grounded_planning_task = grounding.ground(problem=self.problem)
		operators = {operator.name: operator for operator in grounded_planning_task.operators}
		op_names, plan_actions_sequence = self._parse_plan_actions(plan_path)
		previous_state = State(self.problem.initial_state, self.domain)

		self.logger.info("Starting to iterate over the actions sequence.")
		for index, (op_name, action) in enumerate(zip(op_names, plan_actions_sequence)):
			try:
				component = self._create_single_trajectory_component(
					index, action, op_name, operators, previous_state)
				trajectory.append(component)
				previous_state = component.next_state

			except AssertionError:
				error_message = f"The operation {op_name} is not applicable! The failed action - {action.name}"
				self.logger.warning(error_message)
				if should_return_partial_trajectory:
					self.logger.debug(f"Returning partial trajectory since the flag was turned on.")
					return trajectory

				raise AssertionError(error_message)

		return trajectory

	def validate_trajectory(self, trajectory: Trajectory) -> bool:
		"""Validate that the last state in the trajectory is indeed the goal state.

		:param trajectory: the trajectory to validate.
		:return: whether or not the trajectory ends with the goal state.
		"""
		last_component: TrajectoryComponent = trajectory[-1]
		last_state = last_component.next_state
		grounded_goals = grounding.ground(problem=self.problem).goals
		return grounded_goals <= last_state.ground_facts()



"""This module exports the domain object to a domain file."""
from collections import defaultdict
from pathlib import Path
from typing import NoReturn, List, Dict

from pddl.pddl import Domain, Action, Predicate, Type


class DomainExporter:
	"""Class that is able to export a domain to a correct PDDL file."""

	def write_action_preconditions(self, predicates: List[Predicate]) -> str:
		"""Writes the predicates of an action's precondition according to the PDDL file format.

		:param predicates: the predicates that are in the domain's definition.
		:return: the formatted string representing the domain's predicates.
		"""
		formatted_preconditions = "{content}"
		if len(predicates) > 1:
			formatted_preconditions = "(and {content})"

		action_predicates = self._write_positive_predicates(predicates)
		return formatted_preconditions.format(content=" ".join(action_predicates))

	@staticmethod
	def _write_positive_predicates(predicates: List[Predicate]) -> List[str]:
		"""writes positive predicates according to the format of a PDDL file.

		:param predicates: the preconditions / effects of an action.
		:return: the formatted positive predicates.
		"""
		action_predicates = []
		for predicate in predicates:
			predicate_formatted_signature = " ".join([f"{name}" for name, _ in predicate.signature])
			predicate_str = f"({predicate.name} {predicate_formatted_signature})"
			action_predicates.append(predicate_str)

		return action_predicates

	@staticmethod
	def _write_negative_predicates(predicates: List[Predicate]) -> List[str]:
		"""writes negative predicates according to the format of a PDDL file.

		:param predicates: the preconditions / effects of an action.
		:return: the formatted negative predicates.
		"""
		action_predicates = []
		for predicate in predicates:
			predicate_formatted_signature = " ".join([f"{name}" for name, _ in predicate.signature])
			predicate_str = f"(not ({predicate.name} {predicate_formatted_signature}))"
			action_predicates.append(predicate_str)

		return action_predicates

	def write_action_effects(self, add_effects: List[Predicate], delete_effects: List[Predicate]) -> str:
		"""Write the effects of an action according to the PDDL file format.

		:param add_effects: the add effects of an action.
		:param delete_effects: the delete effects of an action.
		:return: the formatted string representing the action's effects.
		"""
		action_effect_predicates = self._write_positive_predicates(add_effects)
		action_effect_predicates += self._write_negative_predicates(delete_effects)
		formatted_effects = "{content}"
		if len(action_effect_predicates) > 1:
			formatted_effects = "(and {content})"

		return formatted_effects.format(content=" ".join(action_effect_predicates))

	def write_action(self, action: Action) -> str:
		"""Write the action formatted string from the action data.

		:param action: The action that needs to be formatted into a string.
		:return: the string format of the action.
		"""
		action_params = " ".join([f"{name} - {types[0]}" for name, types in action.signature])
		action_preconds = self.write_action_preconditions(action.precondition)
		action_effects = self.write_action_effects(action.effect.addlist, action.effect.dellist)
		return f"(:action {action.name}\n" \
			   f"\t:parameters   ({action_params})\n" \
			   f"\t:precondition {action_preconds}\n" \
			   f"\t:effect       {action_effects})\n" \
			   f"\n"

	@staticmethod
	def write_predicates(predicates: Dict[str, Predicate]) -> str:
		"""Writes the predicates formatted according to the domain file format.

		:param predicates: the predicates that are in the domain's definition.
		:return: the formatted string representing the domain's predicates.
		"""
		predicates_str = "(:predicates\n{predicates})\n\n"
		predicates_strings = []
		for predicate_name, predicate in predicates.items():
			predicate_params = " ".join(
				[f"{name} - {types[0]}" for name, types in predicate.signature])
			predicates_strings.append(f"\t({predicate_name} {predicate_params})")

		return predicates_str.format(predicates="\n".join(predicates_strings))

	def write_constants(self, constants: Dict[str, Type]) -> str:
		"""Writes the constants of the domain to the new domain file.

		:param constants: the constants that appear in the domain object.
		:return: the representation of the constants as a canonical PDDL string.
		"""
		constants_str = "(:constants\n{constants})\n\n"
		sorted_consts_types = defaultdict(list)
		for constant_name, constant_type in constants.items():
			sorted_consts_types[constant_type.parent.name].append(constant_name)

		constants_content = self.format_type_like_string(sorted_consts_types)
		return constants_str.format(constants="\n".join(constants_content))

	@staticmethod
	def format_type_like_string(sorted_type_like_objects: Dict[str, List[str]]) -> List[str]:
		"""formats the string that are of the same format as types. This applies to both consts and types.

		:param sorted_type_like_objects: the type like objects that are being formatted into a list of strings.
		:return: the formatted strings as a list.
		"""
		type_like_object_content = []
		for pddl_type_name, sub_types in sorted_type_like_objects.items():
			type_like_object_pddl_str = "\t"
			type_like_object_pddl_str += "\n\t".join([child_type for child_type in sub_types[:-1]])
			type_like_object_pddl_str += f"\n\t{sub_types[-1]} - {pddl_type_name}"
			type_like_object_content.append(type_like_object_pddl_str)

		return type_like_object_content

	def write_types(self, types: Dict[str, Type]) -> str:
		"""Writes the definitions of the types according to the PDDL file format.

		:param types: the types that are available in the learned domain.
		:return: the formatted string representing the types in the PDDL domain file.
		"""
		types_str = "(:types\n{types_content})\n\n"
		sorted_types = defaultdict(list)
		for type_name, pddl_type in types.items():
			if pddl_type.parent is not None:
				sorted_types[pddl_type.parent.name].append(type_name)
			else:
				continue

		types_content = self.format_type_like_string(sorted_types)
		return types_str.format(types_content="\n".join(types_content))

	def export_domain(self, domain: Domain, export_path: Path) -> NoReturn:
		"""Export the domain object to a correct PDDL file.

		:param domain: the learned domain object.
		:param export_path: the path to the file that the domain would be exported to.
		"""
		domain_types = self.write_types(domain.types)
		domain_consts = self.write_constants(domain.constants) if len(domain.constants) > 0 else ""
		domain_headers = f"(define (domain {domain.name})\n" \
						 "(:requirements :strips :typing)\n\n" \
						 f"{domain_types}\n" \
						 f"{domain_consts}" \
						 "{domain_content})"
		domain_content = self.write_predicates(domain.predicates)

		for action in domain.actions.values():
			domain_content += self.write_action(action)

		with open(export_path, "wt") as export_domain_file:
			export_domain_file.write(domain_headers.format(domain_content=domain_content))



from .grounded_action import GroundedAction
from .parameter_binding import ParameterBinding
from .state import State
from .trajectory_component import TrajectoryComponent
from .types import Trajectory, Mode
from .comparable_predicate import ComparablePredicate



from distutils.core import setup
from setuptools import find_packages

setup(name='sam-learner',
      version='1.6.6',
      python_requires=">=3.8",
      description='Safe Action Model Learner',
      author='Argaman Mordoch',
      packages=find_packages(exclude=["tests", "utils"]),
     )


"""Locates actions and grounds them according to the input data."""
from typing import List, Tuple

from pddl.pddl import Action, Domain

from sam_learner.sam_models.parameter_binding import ParameterBinding


def parse_plan_action_string(action_str: str) -> Tuple[str, List[str]]:
	"""Parse the action string and returns its name and parameters.

	:param action_str: the string representing the action. should look like:
	 	(load-truck obj23 tru2 pos2)
	:return: the action's name and it's arguments.
	"""
	action_data = action_str.strip("()").split(" ")
	action_name = action_data[0]
	action_params = action_data[1:]
	if action_params == [""]:
		action_params = []

	return action_name, action_params


def locate_lifted_action(domain: Domain, action_name: str) -> Action:
	"""Locate the input action in the domain.

	:param domain: the problem domain that is being used.
	:param action_name: the name of the action to locate.
	:return: the action object in the domain.
	"""
	try:
		return domain.actions[action_name]

	except KeyError:
		raise ValueError("Action not found in the domain!")


def ground_lifted_action(action: Action, parameters_values: List[str]) -> Tuple[
	Action, List[ParameterBinding]]:
	"""Locate the action's parameters and binds them to grounded values.

	:param action: the action to be grounded.
	:param parameters_values: the values of the action's parameters found in the plan data.
	:return: The action and it's parameters binding.
	"""
	binding = []
	for param_value, action_param in zip(parameters_values, action.signature):
		parameter_name, types = action_param
		binding.append(ParameterBinding(parameter_name, types, param_value))

	return action, binding



"""Module to validate the correctness of the learned action models that were generated."""
import csv
import logging
import os
import re
import signal
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, NoReturn, Dict, List, Any

from heuristics import lm_cut
from pyperplan import search_plan
from search import greedy_best_first_search
from task import Operator

from sam_learner.core import TrajectoryGenerator
from sam_learner.sam_models import Trajectory

PLAN_SEARCH_TIME_LIMIT = int(os.environ.get('SEARCH_TIME', "300"))
VALID_STATUS = "valid plan!"
INVALID_PLAN_STATUS = "invalid plan!"
FAILED_ACTION_NAME_RE = re.compile(r".* - (\w+-?\w*)")


class ValidationResult:
	"""Representation of the plan validation result."""
	status: str
	error: Optional[str]
	failed_action_name: Optional[str]

	def __init__(self, status: str, error: Optional[str] = None, failed_action_name: Optional[str] = None):
		self.status = status
		self.error = error
		self.failed_action_name = failed_action_name


class TimeoutException(Exception): pass


@contextmanager
def time_limit(seconds):
	def signal_handler(signum, frame):
		raise TimeoutException("Timed out!")

	signal.signal(signal.SIGALRM, signal_handler)
	signal.alarm(seconds)
	try:
		yield
	finally:
		signal.alarm(0)


def extract_failed_action_name(error_message) -> str:
	"""Extract the failed action name from the error message.

	:param error_message: the message of raised when trying to create a trajectory.
	:return: the name of the failed action.
	"""
	match = FAILED_ACTION_NAME_RE.match(error_message)
	return match.group(1)


class DomainValidator:
	"""Class that validates the correctness of the domains that were learned by the action model learner.

	The validation process works as follows:
		* Using the learned domain we create plans for the test set problems.
		* For each of the created plans we validate its correctness using the validate URL.
		* We need to validate that the plan that was created is safe, i.e. every action is applicable according to the
			preconditions and effects allowed by the original domain.

	Attributes:
		expected_domain_path: the path to the complete domain containing all of the actions and their preconditions and effects.
	"""

	STATISTICS_COLUMNS_NAMES = ["domain_name", "generated_domain_file_name", "problem_file_name", "plan_generated",
								"plan_validation_status", "validation_error"]

	logger: logging.Logger
	expected_domain_path: str

	def __init__(self, expected_domain_path: str):
		self.logger = logging.getLogger(__name__)
		self.expected_domain_path = expected_domain_path

	def write_plan_file(self, plan_actions: List[Operator], solution_path: Path) -> NoReturn:
		"""Write the plan that was created using the learned domain file.

		:param plan_actions: the actions that were executed in the plan.
		:param solution_path: the path to the file to export the plan to.
		"""
		self.logger.info(f"Writing the plan file - {solution_path}")
		with open(solution_path, "wt") as solution_file:
			solution_file.writelines([f"{operator.name}\n" for operator in plan_actions])

	def generate_test_set_plans(self, tested_domain_file_path: str,
								test_set_directory: str) -> Dict[str, Optional[str]]:
		"""Generate plans for the given domain file and the given test set directory.

		:param tested_domain_file_path: the path to the tested domain in which we want to create actions for.
		:param test_set_directory: the path to the directory containing the test set problems.
		:return: dictionary containing a mapping between the problem paths and their corresponding solutions.
		"""
		successful_plans = {}
		self.logger.info("Solving the test set problems using the learned domain!")
		for problem_file_path in Path(test_set_directory).glob("*.pddl"):
			self.logger.debug(f"Solving problem - {problem_file_path.stem}")
			try:
				with time_limit(PLAN_SEARCH_TIME_LIMIT):
					self.search_plan_with_time_limit(
						problem_file_path, successful_plans, test_set_directory,
						tested_domain_file_path)

			except TimeoutException as e:
				self.logger.info("Solver execution timed out!")
				successful_plans[str(problem_file_path)] = None

		return successful_plans

	def search_plan_with_time_limit(
			self, problem_file_path: Path, successful_plans: Dict[str, Optional[str]],
			test_set_directory: str, tested_domain_file_path: str) -> NoReturn:
		"""Search for a plan for the specific problem with a time limit so that the process won't go on forever.

		:param problem_file_path: the path to the problem file.
		:param successful_plans: the dictionary containing the successful plans.
		:param test_set_directory: the directory containing the problems to solve.
		:param tested_domain_file_path: the domain used to solve the problems.
		"""
		plan_actions = search_plan(tested_domain_file_path, problem_file_path, greedy_best_first_search,
								   lm_cut.LmCutHeuristic)
		if plan_actions is not None and len(plan_actions) > 0:
			self.logger.info(f"Solution Found from problem -{problem_file_path.stem}!")
			solution_path = Path(test_set_directory, f"{problem_file_path.stem}_plan.solution")
			self.write_plan_file(plan_actions, solution_path)
			successful_plans[str(problem_file_path)] = str(solution_path)

		else:
			self.logger.info(f"Could not find a solution for problem - {problem_file_path.stem}")
			successful_plans[str(problem_file_path)] = None

	def validate_plan(self, problem_path: str, plan_path: str) -> ValidationResult:
		"""Validate the correctness of the learned domain against the domain learned using the learner algorithm.

		:param problem_path: the to the test set problem.
		:param plan_path: the path to the plan generated by a solver.
		:return: an object representing the validation status of the plan.
		"""
		self.logger.info(f"Validating the plan generated for the problem - {problem_path}")
		trajectory_generator = TrajectoryGenerator(self.expected_domain_path, problem_path)
		try:
			trajectory_generator.generate_trajectory(plan_path)
			return ValidationResult(VALID_STATUS)

		except AssertionError as error:
			self.logger.warning(f"The plan received is not applicable! {error}")
			# Extracting the failed action name
			error_message = str(error)
			failed_action_name = extract_failed_action_name(error_message)
			return ValidationResult(INVALID_PLAN_STATUS, str(error), failed_action_name)

	def extract_applicable_plan_components(self, problem_path: str, plan_path: str) -> Trajectory:
		"""Extract the partial trajectory from the failed plan.

		:param problem_path: the path to the problem file.
		:param plan_path: the path to the failed plan.
		:return: the partial applicable trajectory.
		"""
		self.logger.info(f"Extracting the applicable trajectory from the plan file - {plan_path}")
		trajectory_generator = TrajectoryGenerator(self.expected_domain_path, problem_path)
		return trajectory_generator.generate_trajectory(plan_path, should_return_partial_trajectory=True)

	def write_statistics(self, statistics: List[Dict[str, Any]], output_statistics_path: str) -> NoReturn:
		"""Writes the statistics of the learned model into a CSV file.

		:param statistics: the object containing the statistics about the learning process.
		:param output_statistics_path: the path to the output file.
		"""
		with open(output_statistics_path, 'wt', newline='') as csv_file:
			writer = csv.DictWriter(csv_file, fieldnames=self.STATISTICS_COLUMNS_NAMES)
			writer.writeheader()
			for data in statistics:
				writer.writerow(data)

	@staticmethod
	def clear_plans(created_plans: List[str]) -> NoReturn:
		"""Clears the directory from the plan files so that the next iterations could continue.

		:param created_plans: the paths to the plans that were created in this iteration.
		"""
		for plan_path in created_plans:
			if plan_path is not None:
				os.remove(plan_path)

	def on_validation_success(
			self, output_statistics_path: str,
			test_set_statistics: List[Dict[str, Any]]) -> NoReturn:
		"""Write the needed statistics if the plans were all validated and were approved.

		:param output_statistics_path: the path to the output file.
		:param test_set_statistics: the statistics object containing the data about the learning process.
		"""
		self.logger.info("All plans are valid!")
		self.write_statistics(test_set_statistics, output_statistics_path)
		self.logger.info("Done!")

	def log_model_safety_report(
			self, domain_name: str, generated_domains_directory_path: str, test_set_directory_path: str,
			output_statistics_path: str) -> NoReturn:
		"""The main entry point that runs the validation process for the plans that were generated using the learned
			domains.

		:param domain_name: the name of the domain that is being validated.
		:param generated_domains_directory_path: the path to the directory containing the generated domains.
		:param test_set_directory_path: the directory containing the test set problems.
		:param output_statistics_path: the path to the output statistics file.
		"""
		test_set_statistics = []
		for generated_domain_path in Path(generated_domains_directory_path).glob("*.pddl"):

			domain_successful_plans = self.generate_test_set_plans(
				str(generated_domain_path), test_set_directory_path)
			plan_paths = [plan for plan in domain_successful_plans.values()]
			validated_plans = []
			for problem_path, plan_path in domain_successful_plans.items():
				domain_statistics = {
					"domain_name": domain_name,
					"generated_domain_file_name": generated_domain_path.stem,
					"problem_file_name": Path(problem_path).stem}

				if plan_path is None:
					domain_statistics["plan_generated"] = False
					continue

				validation_status = self.validate_plan(problem_path, plan_path)
				domain_statistics["plan_generated"] = True
				domain_statistics["plan_validation_status"] = validation_status.status
				domain_statistics["validation_error"] = validation_status.error
				validated_plans.append(validation_status.status == VALID_STATUS)
				test_set_statistics.append(domain_statistics)

			self.clear_plans(plan_paths)
			if all(validated_plans):
				self.on_validation_success(output_statistics_path, test_set_statistics)
				return

		self.write_statistics(test_set_statistics, output_statistics_path)
		self.logger.info("Done!")


if __name__ == '__main__':
	try:
		logging.basicConfig(level=logging.DEBUG)
		args = sys.argv
		validator = DomainValidator(expected_domain_path=args[1])
		validator.log_model_safety_report(
			domain_name=args[2],
			generated_domains_directory_path=args[3],
			test_set_directory_path=args[4],
			output_statistics_path=args[5])

	except Exception as e:
		print(e)



"""This module creates plans using random walk steps. This will extend the amount of actions."""
import logging
import random
import sys

from pathlib import Path

import grounding
from pddl.parser import Parser
from pddl.pddl import Domain, Problem
from typing import NoReturn, List

from task import Task

from sam_learner.sam_models.state import State


class RandomWalkPlansGenerator:
    """Class that generates plans for a domain and a problem using random walk algorithm.

    Attributes:
        logger: the logger of the class.
        domain: the PDDL domain data.
        problem: the PDDL problem data.
    """

    logger: logging.Logger
    domain: Domain
    problem: Problem
    problem_path: Path
    output_directory_path: Path

    def __init__(self, domain_path: str, problem_path: str, output_directory_path: str):
        parser = Parser(domain_path, problem_path)
        self.logger = logging.getLogger(__name__)
        self.domain = parser.parse_domain(read_from_file=True)
        self.problem = parser.parse_problem(dom=self.domain, read_from_file=True)
        self.problem_path = Path(problem_path)
        self.output_directory_path = Path(output_directory_path)

    def generate_single_plan(self, task: Task, max_plan_steps: int) -> List[str]:
        """Generate a plan from a randomly selected applicable actions (the random walk).

        :param task: the task that generates the possible transitions from the initial state.
        :param max_plan_steps: the maximal length of the generated plans.
        :return: a list containing the action sequences that were generated.
        """
        if max_plan_steps <= 0:
            raise ValueError("Given illegal value of steps for the plan!")
        self.logger.info(f"Starting to generate a plan for the domain - {self.domain.name} and "
                         f"the problem - {self.problem.name}")
        actions_sequence = []
        current_state = State(self.problem.initial_state, self.domain).ground_facts()
        num_steps = 0
        while num_steps < max_plan_steps:
            possible_transitions = task.get_successor_states(current_state)
            if len(possible_transitions) == 0:
                return actions_sequence

            action, state = random.choice(possible_transitions)
            self.logger.debug(f"generated the applicable action {action}")
            actions_sequence.append(f"{action.name}\n")
            current_state = state
            num_steps += 1

        actions_sequence[-1] = actions_sequence[-1].strip("\n")
        return actions_sequence

    def generate_plans(self, max_plan_steps: int, num_plans: int) -> List[List[str]]:
        """Generate plans with maximal length given as input.

        :param max_plan_steps: the maximal length of the output plans.
        :param num_plans: the number of plans to generate.
        :returns the plans as action sequences.
        """
        grounded_planning_task: Task = grounding.ground(problem=self.problem)
        plans = []
        while len(plans) < num_plans:
            plan = self.generate_single_plan(grounded_planning_task, max_plan_steps)
            if len(plan) == 0:
                continue

            plans.append(plan)

        return plans

    def export_plans(self, generated_plans: List[List[str]], plan_length: int) -> NoReturn:
        """Export the plans to plan file according to the correct format.

        :param generated_plans: the plans that were generated using random walk.
        :param plan_length: the length of the generated plans.
        """
        for index, plan in enumerate(generated_plans, 1):
            with open(f"{str(self.output_directory_path / self.problem_path.stem)}_plan_"
                      f"{index}_max_len_{plan_length}.solution", "wt") as plan_file:
                plan_file.writelines(plan)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    directory_path = sys.argv[1]
    domain_path = sys.argv[2]
    problem_files_glob = sys.argv[3]
    for file_path in Path(directory_path).glob(f"{problem_files_glob}*.pddl"):
        print(f"working on - {file_path}")
        gen = RandomWalkPlansGenerator(domain_path,
                                       file_path,
                                       directory_path)

        plans = gen.generate_plans(max_plan_steps=30, num_plans=5)
        gen.export_plans(generated_plans=plans, plan_length=30)



"""Contains the new types for the action model learner."""
from typing import List, Literal

from sam_learner.sam_models.trajectory_component import TrajectoryComponent

Trajectory = List[TrajectoryComponent]
Mode = Literal["production", "development"]
